//------------------------------------------------------------------------------
//	$Id: scubaquarium.js,v 1.3 2018/03/08 19:35:53 wolf Exp $
//------------------------------------------------------------------------------
//	This program is free software.
//	It comes without any warranty, to the extent permitted by applicable law.
//	You can redistribute it and/or modify it under the terms of the
//		"Do What The Fuck You Want To" Public License, Version 2,
//	as published by Sam Hocevar. See http://www.wtfpl.net/ for more details.
//------------------------------------------------------------------------------

function addListener (element, eventName, handler) {

	if (element.addEventListener) {
		element.addEventListener(eventName, handler, false);
	} else if (window.document.attachEvent) {
		element.attachEvent('on'+eventName, handler);
	} else {
		window.alert ('Neither addEventListener nor attachEvent defined.');
	}
}

addListener(window, 'load', function() {

	function createAquarium (container) {

		//--------------------------------------------------------------
		//	clear container
		//--------------------------------------------------------------

		while (container.firstChild) container.removeChild(container.firstChild);

		//--------------------------------------------------------------
		//	auxiliary methodts for getting parameter
		//--------------------------------------------------------------

		function getIntParam (name, minValue, maxValue) {

			var value = parseInt (container.getAttribute('data-' + name), 10) || 0;
			return Math.max (minValue, Math.min (maxValue, value));
		}

		function getFactorParam (name, minValue, maxValue) {

			var factor = parseFloat (container.getAttribute('data-' + name)) || 0.5;
			return Math.max (minValue, Math.round (Math.min(1,factor) * maxValue));
		}

		//--------------------------------------------------------------
		//	get parameter
		//--------------------------------------------------------------

		var nFishTypes = getIntParam ('n-fish-types', 1, 1000);
		var nWormTypes = getIntParam ('n-worm-types', 1, 1000);
		var prefix     = container.getAttribute('data-tile-name-prefix');
		var suffix     = container.getAttribute('data-tile-name-suffix');
		var tileWidth  = getIntParam ('tile-width', 64, 256);
		var tileHeight = getIntParam ('tile-height', 64, 256);
		var borderSize = getIntParam ('border-size', 0, 24);

		if (!prefix || !suffix) return;

		//--------------------------------------------------------------
		//	adapt to window size
		//--------------------------------------------------------------

		var maxWidth  = container.clientWidth - 2 * borderSize;
		var maxHeight = window.innerHeight    - 2 * borderSize;

		if ('getComputedStyle' in window) {

			var cs = window.getComputedStyle(container);
			maxWidth -= parseFloat(cs.getPropertyValue('border-left-width')) || 0;
			maxWidth -= parseFloat(cs.getPropertyValue('padding-left')) || 0;
			maxWidth -= parseFloat(cs.getPropertyValue('padding-right')) || 0;
			maxWidth -= parseFloat(cs.getPropertyValue('border-right-width')) || 0;
		}

		var nRows = getIntParam ('n-rows'   , 2, Math.floor(maxHeight/ tileHeight));
		var nCols = getIntParam ('n-columns', 2, Math.floor(maxWidth / tileWidth ));
		var nFish = getFactorParam ('fish-factor', 1, (nRows-1) * nCols - 1);

		//--------------------------------------------------------------
		//	create fish urls
		//--------------------------------------------------------------

		var noFishURL = prefix + 'fish' + 0 + suffix;
		var fishTypes = [];

		for (var fishTypeIndex=0; fishTypeIndex < nFishTypes; fishTypeIndex++) {

			fishTypes[fishTypeIndex] = {
				'leftURL':  prefix + 'fish' + (fishTypeIndex+1) + 'l' + suffix,
				'rightURL': prefix + 'fish' + (fishTypeIndex+1) + 'r' + suffix
			};
		}

		//--------------------------------------------------------------
		//	create worm urls
		//--------------------------------------------------------------

		var noWormURL = prefix + 'worm' + 0 + suffix;
		var wormTypes = [];

		for (var wormTypeIndex=0; wormTypeIndex < nWormTypes; wormTypeIndex++) {

			wormTypes[wormTypeIndex] = {
				'outURL':  prefix + 'worm' + (wormTypeIndex+1) + 'o' + suffix
			};
		}

		//--------------------------------------------------------------
		//	aux method: draw url at row col
		//--------------------------------------------------------------

		var images = [];

		function drawAtRowColURL (row, col, url) {

			images[row][col].setAttribute('src', url);
		}

		//--------------------------------------------------------------
		//	draw fish
		//--------------------------------------------------------------

		function drawFish (fish) {

			var url = fish.dir > 0 ? fish.type.rightURL : fish.type.leftURL;
			images[fish.row][fish.col].setAttribute('src', url);
		}

		//--------------------------------------------------------------
		//	fish indices
		//--------------------------------------------------------------

		var fishByRowCol = [];	// fish by row, col, null = empty
		var fishById     = [];	// 0..nFish-1

		for (var r=0; r<nRows; r++) {

			fishByRowCol.push (new Array (nCols));
		}

		//--------------------------------------------------------------
		//	populate aquarium with fishes
		//--------------------------------------------------------------

		function populateWithFishes() {

			//------------------------------------------------------
			//	possible positions
			//------------------------------------------------------

			var positionsCol = [];
			var positionsRow = [];

			for (var row=1; row<nRows; row++) {

				for (var col=0; col<nCols; col++) {

					positionsCol.push(col);
					positionsRow.push(row);
				}
			}

			//------------------------------------------------------
			//	possible fish types
			//------------------------------------------------------

			var myFishTypes = fishTypes.slice();

			//------------------------------------------------------
			//	populate
			//------------------------------------------------------

			for (var fishId=0; fishId<nFish; fishId++) {

				//----------------------------------------------
				//	out of positions -> done
				//----------------------------------------------

				if (!positionsCol.length) break;

				var posIndex = Math.floor (Math.random() * positionsCol.length);
				col = positionsCol[posIndex];
				row = positionsRow[posIndex];
				positionsCol.slice(posIndex, 1);
				positionsRow.slice(posIndex, 1);

				//----------------------------------------------
				//	out of types → recycle
				//----------------------------------------------

				if (!myFishTypes.length) myFishTypes = fishTypes.slice();

				var fishTypeIndex = Math.floor (Math.random() * myFishTypes.length);
				var type          = myFishTypes[fishTypeIndex];
				myFishTypes.slice(fishTypeIndex, 1);

				//----------------------------------------------
				//	create fish
				//----------------------------------------------

				var fish = {
					'id':   fishId,
					'type': type,
					'col':  col,
					'row':  row,
					'dir':  Math.floor(Math.random(2)) * 2 - 1
				};

				//----------------------------------------------
				//	register fish
				//----------------------------------------------

				fishById[fishId] = fish;
				fishByRowCol[row][col] = fish;

				//----------------------------------------------
				//	and draw fish
				//----------------------------------------------

				drawFish (fish);
			}
		}

		//--------------------------------------------------------------
		//	worms hide if a fish is nearby
		//--------------------------------------------------------------

		function scareWorm (col) {

			if (col < 0 || col >= nCols) return;
			drawAtRowColURL (0, col, noWormURL);
		}

		//--------------------------------------------------------------
		//	animate fishes and worms
		//--------------------------------------------------------------

		function animate () {

			//------------------------------------------------------
			//	animate fish
			//------------------------------------------------------

			var fishId = Math.floor (Math.random() * fishById.length);
			var fish = fishById[fishId];

			//------------------------------------------------------
			//	old position
			//------------------------------------------------------

			var col = fish.col;
			var row = fish.row;
			var dir = fish.dir;

			//------------------------------------------------------
			//	compute new position
			//------------------------------------------------------

			if (col <= 0 && dir < 0) {
				dir=1;
			} else if (col>=nCols-1 && dir>0) {
				dir=-1;
			} else if (!fishByRowCol[row][col+dir] && Math.random() < 0.9) {
				col+=dir;
			} else if (row > 1 && !fishByRowCol[row-1][col+dir] && Math.random() < 0.5) {
				col+=dir;
				row-=1;
			} else if (row < nRows-1 && !fishByRowCol[row+1][col+dir]) {
				col+=dir;
				row++;
			} else if (row > 1 && !fishByRowCol[row-1][col+dir]) {
				col+=dir;
				row-=1;
			} else if (!fishByRowCol[row][col+dir]) {
				col+=dir;
			} else if (Math.random()<0.3) {
				dir = -dir;
			}

			//------------------------------------------------------
			//	remove fish from old pos if position changed 
			//------------------------------------------------------

			if (row != fish.row || col != fish.col) {

				drawAtRowColURL (fish.row, fish.col, noFishURL);

				fishByRowCol[fish.row][fish.col] = null;
				fishByRowCol[row][col] = fish;

				//----------------------------------------------
				//	scare worms
				//----------------------------------------------

				if (row==1) {

					scareWorm(col-1);
					scareWorm(col  );
					scareWorm(col+1);
				}
			}

			//------------------------------------------------------
			//	store new position and draw at new position
			//------------------------------------------------------

			if (row != fish.row || col != fish.col || dir != fish.dir) {

				fish.row = row;
				fish.col = col;
				fish.dir = dir;

				drawFish(fish);
			}

			//------------------------------------------------------
			//	worms show if no fish nearby and hide at random
			//------------------------------------------------------

			var wormCol = Math.floor (Math.random() * nCols);

			if (Math.random() < 0.1) {

				drawAtRowColURL (0, wormCol, noWormURL);

			} else if (!fishByRowCol[1][wormCol] &&
				(wormCol<=0       || !fishByRowCol[1][wormCol-1]) &&
				(wormCol>=nCols-1 || !fishByRowCol[1][wormCol+1])) {

				var wormType = wormTypes[Math.floor(Math.random() * wormTypes.length)];
				drawAtRowColURL (0, wormCol, wormType.outURL);
			}
		}

		//--------------------------------------------------------------
		//	scare worms on mouseover or touch
		//--------------------------------------------------------------

		function onScareWorm(evt) {

			scareWorm(parseInt(this.getAttribute('data-col'), 10));
		}

		//--------------------------------------------------------------
		//	create and populate aquarium
		//--------------------------------------------------------------

		var table = document.createElement('table');
		var tbody = document.createElement('tbody');

		table.style.cellSpacing = 0;
		table.style.borderWidth  = borderSize + 'px';
		table.className = 'scubaquarium';
		table.appendChild(tbody);

		for (var row=nRows-1; row>=0; row--) {

			var tr = document.createElement('tr');
			var imagesInRow = [];

			for (var col=0; col<nCols; col++) {

				var td  = document.createElement('td');
				var img = document.createElement('img');

				img.setAttribute('src', row>0 ? noFishURL : noWormURL);
				img.style.width  = tileWidth  + 'px';
				img.style.height = tileHeight + 'px';

				if (row===0) {

					img.setAttribute('data-col', col);
					addListener (img, 'mouseover', onScareWorm);
					addListener (img, 'touchstart', onScareWorm);
				}

				td.appendChild(img);
				tr.appendChild(td);

				imagesInRow.push(img);
			}

			images[row] = imagesInRow;

			tbody.appendChild(tr);
		}

		container.appendChild(table);

		//--------------------------------------------------------------
		//	put fishes into the aquarium
		//--------------------------------------------------------------

		populateWithFishes();

		//--------------------------------------------------------------
		//	start fish and worm animation
		//--------------------------------------------------------------

		window.setInterval(animate, 300);
	}

	//----------------------------------------------------------------------
	//	create an aquarium for each div.aquarium
	//----------------------------------------------------------------------

	function createAll () {

		var containers = document.getElementsByClassName('scubaquarium');
		for (var c=0; c<containers.length; c++) {

			var container = containers[c];

			if (container.tagName.toLowerCase() != 'div') continue;
			createAquarium(container);
		}
	}

	//----------------------------------------------------------------------
	//	create an aquarium for each div.aquarium on window resize
	//----------------------------------------------------------------------

	addListener(window, 'resize', createAll);

	//----------------------------------------------------------------------
	//	create an aquarium for each div.aquarium on page load
	//----------------------------------------------------------------------

	createAll();
});

//------------------------------------------------------------------------------
//	$Id: scubaquarium.js,v 1.3 2018/03/08 19:35:53 wolf Exp $
//------------------------------------------------------------------------------
